"""Generate site in output directory."""

from pathlib import Path
from shutil import rmtree
from sqlite3 import Connection
from typing import Callable

from .page import generate_complete_html as generate_index


data = Path(__file__).parent/"data"

Generator = Callable[[Path], None]


class OutputDirectory:
    """Represents output directory of generated site."""
    def __init__(self, path: Path):
        self.path = path

    def clear(self) -> None:
        """Clear contents of output directory."""
        assert not self.path.exists() or self.path.is_dir()
        rmtree(self.path, ignore_errors=True)
        self.path.mkdir()

    def generate(self, *args: Generator) -> None:
        """Run generators."""
        self.clear()
        for arg in args:
            arg(self.path)


class IndexGenerator:
    """Generates index.html."""
    def __init__(self, con: Connection, options: str, title: str = "Slipbox"):
        self.con = con
        self.options = options
        self.title = title

    def run(self, out: Path) -> None:
        """Generate index.html inside output directory."""
        generate_index(self.con, self.options, out, self.title)


def copy(source: Path, dest: Path) -> None:
    """Copy text from source to dest Path."""
    dest.write_text(source.read_text())


def generate_js(out: Path) -> None:
    """Generate app.js inside output directory."""
    copy(data/"app.js", out/"app.js")


def generate_css(out: Path) -> None:
    """Generates CSS files: style.css and base.css."""
    copy(data/"style.css", out/"style.css")
    copy(data/"base.css", out/"base.css")


class ImagesGenerator:
    """Copies images from database into output directory."""
    def __init__(self, con: Connection):
        self.con = con

    def run(self, out: Path) -> None:
        """Copy images into output directory."""
        images = out/"images"
        images.mkdir()
        sql = "SELECT filename, binary FROM Images"
        for filename, binary in self.con.execute(sql):
            image = images/filename
            image.write_bytes(binary)


def main(con: Connection,
         options: str,
         out: Path,
         title: str = "Slipbox") -> None:
    """Generate all files."""
    OutputDirectory(out).generate(
        IndexGenerator(con, options, title).run,
        ImagesGenerator(con).run,
        generate_js,
        generate_css,
    )



"""Process notes and store results in database.

Preprocessors insert metadata blocks into notes as code blocks.
They are mainly used to pass note metadata to Pandoc filters even when
the input is the concatenation of several files.
"""

from configparser import ConfigParser
from pathlib import Path
from sqlite3 import Connection
import sys
from typing import Any, Sequence

from . import utils
from .batch import Batch
from .data import process_csvs
from .scan import build_command
from .secparse import parse_sections, SectionParser


DOKUWIKI_TEMPLATE = """
<code>
[slipbox-metadata]
{}
</code>
"""

# NOTE Some brackets are escaped.
LATEX_TEMPLATE = r"""
\begin{{verbatim}}
[slipbox-metadata]
{}
\end{{verbatim}}
"""

MARKDOWN_TEMPLATE = """
```
[slipbox-metadata]
{}
```
"""

MEDIAWIKI_TEMPLATE = """
<pre>[slipbox-metadata]
{}</pre>
"""

ORG_TEMPLATE = """
#+begin_example
[slipbox-metadata]
{}
#+end_example
"""

RST_TEMPLATE = """
.. code::

    [slipbox-metadata]
    {}

"""

T2T_TEMPLATE = MARKDOWN_TEMPLATE

TEXTILE_TEMPLATE = """
bc. [slipbox-metadata]
{}

"""

METADATA_TEMPLATES = {
    ".dokuwiki": DOKUWIKI_TEMPLATE,
    ".latex": LATEX_TEMPLATE,
    ".markdown": MARKDOWN_TEMPLATE,
    ".md": MARKDOWN_TEMPLATE,
    ".mdown": MARKDOWN_TEMPLATE,
    ".org": ORG_TEMPLATE,
    ".rst": RST_TEMPLATE,
    ".t2t": T2T_TEMPLATE,
    ".tex": LATEX_TEMPLATE,
    ".textile": TEXTILE_TEMPLATE,
    ".wiki": MEDIAWIKI_TEMPLATE,
}


def render_metadata(template: str, **fields: Any) -> str:
    """Render metadata code block using template."""
    body = '\n'.join(
        f"{k}={v}"
        for k, v in fields.items()
    )
    return template.format(body)


def preprocess(template: str, *sources: Path, basedir: Path) -> str:
    """Preprocess notes (sources) by inserting code blocks generated
    using the template.
    """
    return "".join(
        render_metadata(template, filename=str(source.relative_to(basedir)))
        + source.read_text(encoding="utf-8")
        for source in sources
    )


def store_html(conn: Connection,
               html: str,
               sources: Sequence[Path]) -> None:
    """Insert HTML sections into Notes table."""
    if not html.strip() or not sources:
        return
    cur = conn.cursor()

    def callback(this: SectionParser) -> None:
        """Callback for parse_sections."""
        cur.execute(
            "UPDATE Notes SET html = ? WHERE id = ?",
            (this.section, this.id_)
        )
    parse_sections(html, callback)


def create_preprocessed_input(
    tempdir: Path,
    batch: Batch,
    basedir: Path
) -> Path:
    """Create preprocessed input to be passed to Pandoc."""
    template = METADATA_TEMPLATES.get(batch.extension, MARKDOWN_TEMPLATE)
    path = tempdir/("input" + batch.extension)
    path.write_text(
        preprocess(template, *batch.paths, basedir=basedir),
        encoding="utf-8"
    )
    return path


def process_batch(conn: Connection,
                  batch: Batch,
                  config: ConfigParser,
                  basedir: Path) -> None:
    """Process batch of input notes."""
    with utils.temporary_directory() as tempdir:
        preprocessed_input = create_preprocessed_input(tempdir, batch, basedir)
        html = tempdir/"temp.html"
        cmd = build_command(preprocessed_input, str(html), basedir,
                            config.get("slipbox", "content_options"))
        retcode = utils.run_command(cmd, cwd=tempdir)
        if retcode:
            print("Scan failed.", file=sys.stderr)
            return
        process_csvs(conn, tempdir)
        store_html(conn, html.read_text(encoding="utf-8"), batch.paths)
        conn.commit()



# type: ignore
"""Test initializer.py."""

from .initializer import DotSlipbox


def test_initialize(tmp_path):
    """Must initialize .slipbox in the parent directory."""
    dot = DotSlipbox(tmp_path).path
    assert dot.is_dir()
    assert dot.joinpath("data.db").is_file()
    assert dot.joinpath("config.cfg").is_file()
    assert dot.joinpath("patterns").is_file()


def test_dot_slipbox_locate(tmp_path):
    """find_dot_slipbox must look for .slipbox in parent directories."""
    assert DotSlipbox.locate(tmp_path) is None

    dot = DotSlipbox(tmp_path)
    assert DotSlipbox.locate(tmp_path).path == dot.path

    child = tmp_path/"child"
    child.mkdir()
    assert DotSlipbox.locate(child).path == dot.path

    for file in dot.path.iterdir():
        file.unlink()
    dot.path.rmdir()
    assert DotSlipbox.locate(child) is None
    assert DotSlipbox.locate(tmp_path) is None


def test_dot_slipbox_patterns(tmp_path):
    """Test patterns property."""
    dot = DotSlipbox(tmp_path)
    patterns = sorted(dot.patterns)
    assert "*.markdown" in patterns
    assert "*.md" in patterns
    assert "*.mdown" in patterns

    patterns = ["*.rst", "*.txt"]
    dot.patterns = patterns
    assert dot.patterns == patterns



"""Look for files that must be compiled."""

import fnmatch
import os
from pathlib import Path
import shlex
from sqlite3 import Connection
from typing import Iterable

from . import utils


def is_recently_modified(timestamp: float, path: Path) -> bool:
    """Check if file has been modified after the timestamp."""
    return path.exists() and os.path.getmtime(path) >= timestamp


def is_file_in_db(path: Path, conn: Connection) -> bool:
    """Check if file is recorded in the database."""
    cur = conn.cursor()
    sql = "SELECT filename FROM Files WHERE filename = ?"
    for _ in cur.execute(sql, (str(path),)):
        return True
    return False


def has_valid_pattern(path: Path,
                      patterns: Iterable[str],
                      basedir: Path) -> bool:
    """Check if path matches one of the patterns."""
    for pattern in patterns:
        relpath = str(path.relative_to(basedir.resolve()))
        if fnmatch.fnmatch(relpath, pattern):
            return True
    return False


def build_command(input_: Path,
                  output: str,
                  basedir: Path,
                  options: str = "") -> str:
    """Construct a single pandoc command to run on input."""
    assert input_.exists()
    data = Path(__file__).parent/"data"
    lua_filter = shlex.quote(str((data/"filter.lua").resolve()))
    cmd = f"{utils.pandoc()} {options} -L{lua_filter} --section-divs " \
        f" -Mlink-citations:true " \
        "--resource-path {} -o {} --extract-media=images".format(
            shlex.quote(str(basedir.resolve())),
            shlex.quote(output),
        )
    return cmd + ' ' + shlex.quote(str(input_.resolve()))



# type: ignore
"""Test processor.py."""
from .processor import preprocess, MARKDOWN_TEMPLATE


def test_preprocess_markdown_with_sources(files_abc, tmp_path):
    """There must be a code block between each file section in the
    result.
    """
    _, *sources = files_abc
    sources[0].write_text("B")
    sources[1].write_text("C")

    content = preprocess(MARKDOWN_TEMPLATE, *sources, basedir=tmp_path)
    template = """
```
[slipbox-metadata]
filename={}
```
"""

    assert template.format(str(sources[0].relative_to(tmp_path))) in content
    assert template.format(str(sources[1].relative_to(tmp_path))) in content


def test_preprocess_markdown_with_no_sources(tmp_path):
    """It must return an empty string."""
    content = preprocess(MARKDOWN_TEMPLATE, basedir=tmp_path)
    assert not content



"""Basic templates.

Also includes functions for building HTML elements.
"""

from pathlib import Path
from string import Template as StringTemplate
from textwrap import indent
from typing import Any, Union


class Template(StringTemplate):
    """Extended string.Template with file loader."""
    @staticmethod
    def load(path: Path) -> "Template":
        """Load template from file."""
        return Template(path.read_text())

    def render(self, **kwargs: Any) -> str:
        """Substitute template variables."""
        return self.safe_substitute(**kwargs)


def render_template(path: Union[str, Path], **kwargs: Any) -> str:
    """Render template from templates directory."""
    basedir = Path(__file__).parent/"templates"
    template = Template.load(basedir/path)
    return template.render(**kwargs)


class Elem:
    """Represents an HTML element."""
    def __init__(self,
                 tag: str,
                 *children: Union[str, "Elem"],
                 **attributes: Any):
        self.tag = tag
        self.children = list(children)
        self.attributes = attributes.copy()


def render(elem: Union[str, Elem], prefix: str = "") -> str:
    """Render HTML from Elem or str."""
    if isinstance(elem, str):
        return prefix + elem
    assert isinstance(elem, Elem)
    attrs = ""
    for key, val in elem.attributes.items():
        attrs += f" {key}={val!r}"
    if not elem.children:
        return "<{tag}{attrs}></{tag}>".format(tag=elem.tag, attrs=attrs)
    children = indent("\n".join(render(child) for child in elem.children),
                      prefix + "  ")
    return "<{tag}{attrs}>\n{children}\n</{tag}>".format(
        tag=elem.tag,
        attrs=attrs,
        children=children)



"""Generate complete HTML containing all slipbox notes."""

from pathlib import Path
import shlex
from sqlite3 import Connection
import subprocess
from typing import Iterable

from .templates import Elem, render, render_template
from .utils import pandoc, temporary_directory


def render_dummy(title: str) -> str:
    """Render dummy markdown."""
    return fr"""---
title: {title}
...

::: {{style="display:none"}}
$\,$
```c
```
:::
"""


def data_path(filename: str) -> Path:
    """Return path in data directory."""
    return Path(__file__).parent/"data"/filename


def data_shell_path(filename: str) -> str:
    """Return shell-escaped filename in data directory."""
    return shlex.quote(str(data_path(filename)))


def generate_active_htmls(conn: Connection) -> Iterable[str]:
    """Get HTML stored in the database for active sections."""
    sql = "SELECT html FROM Notes WHERE html IS NOT NULL ORDER BY id ASC"
    return (html.strip() for html, in conn.execute(sql))


def generate_data(conn: Connection) -> Iterable[str]:
    """Generate slipbox data in javascript."""
    yield "for(const sec of document.querySelectorAll('section.slipbox-note'))"
    yield "{"
    yield "  window.slipbox.addNote(sec.id)"
    yield "}"
    sql = "SELECT src, dest, tag FROM ValidLinks ORDER BY src, dest, tag"
    for src, dest, tag in conn.execute(sql):
        tag = "null" if not tag else repr(tag)
        yield f"window.slipbox.addLink({str(src)!r}, {str(dest)!r}, {tag})"


def generate_javascript(conn: Connection) -> Iterable[str]:
    """Generate slipbox javascript code."""
    yield '<script type="module" src="app.js" defer></script>'
    yield '<script>'
    yield "window.addEventListener('DOMContentLoaded', () => {"
    yield from generate_data(conn)
    yield "window.initSlipbox()"
    yield "})"
    yield "</script>"


def create_bibliography(conn: Connection) -> str:
    """Create bibliography HTML section from database entries."""
    sql = "SELECT key, text FROM Bibliography ORDER BY key"
    items = '\n'.join(
        render_template("bibliography__item.html", **dict(
            href=f"#{key}", term=f"[@{key[4:]}]", description=text
        ))
        for key, text in conn.execute(sql)
    )
    return render_template("bibliography.html", items=items)


def create_tags(conn: Connection) -> str:
    """Create HTML section that lists all tags.

    Also list untagged notes.
    """
    rows = conn.execute(
        "SELECT tag, COUNT(*) FROM Tags GROUP BY tag ORDER BY tag"
    )
    items = (Elem("li", Elem("a", tag, href=f"#tags/{tag[1:]}"), f" ({count})")
             for tag, count in rows)
    section = Elem("section",
                   Elem("h1", "Tags"),
                   Elem("ul", *items),
                   id="tags",
                   title="Tags",
                   **{"class": "level1"})
    untagged = list(conn.execute("SELECT * FROM Untagged"))
    if untagged:
        section.children.append(Elem("h2", "Untagged notes"))
        items = (Elem("li", value=str(nid)) for nid, in untagged)
        section.children.append(Elem("ol", *items,
                                     **{"class": "slipbox-list"}))
    return render(section)


def create_tag_page(conn: Connection, tag: str) -> str:
    """Create HTML section that lists all notes with the tag."""
    sql = """
        SELECT id FROM Tags NATURAL JOIN Notes WHERE tag = ? ORDER BY id
    """
    items = []
    for nid, in conn.execute(sql, (tag,)):
        item = Elem("li", value=str(nid))
        items.append(item)
    section = Elem("section",
                   Elem("h1", tag),
                   Elem("ol", *items, **{"class": "slipbox-list"}),
                   id=f"tags/{tag[1:]}",
                   title=tag,
                   **{"class": "level1"})
    return render(section)


def create_tag_pages(conn: Connection) -> str:
    """Create all tag pages."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    return '\n'.join(create_tag_page(conn, tag) for tag in tags)


def create_reference_page(conn: Connection, reference: str) -> str:
    """Create HTML section that lists all notes that cite the reference."""
    sql = """
        SELECT note, text FROM Citations
            JOIN Notes ON Citations.note = Notes.id
                JOIN Bibliography ON Bibliography.key = Citations.reference
                    WHERE reference = ?
                        ORDER BY note
    """
    items = []
    text = ""
    for note, _text in conn.execute(sql, (reference,)):
        text = _text
        item = Elem("li", value=str(note))
        items.append(item)
    section = Elem("section",
                   Elem("h1", '@' + reference[4:]),
                   Elem("p", text),
                   Elem("ol", *items, **{"class": "slipbox-list"}),
                   id=reference,
                   title=reference,
                   **{"class": "level1"})
    return render(section)


def create_reference_pages(conn: Connection) -> str:
    """Create all reference pages."""
    rows = conn.execute("SELECT key FROM Bibliography ORDER BY key")
    references = (row[0] for row in rows)
    return '\n'.join(create_reference_page(conn, ref) for ref in references)


def generate_complete_html(conn: Connection,
                           options: str,
                           out: Path,
                           title: str = "Slipbox") -> None:
    """Create final HTML file with javascript."""
    with temporary_directory() as tempdir:
        (tempdir/"nav.html").write_text(render_template("nav.html",
                                                        title=title))
        script = tempdir/"script.js"
        html = tempdir/"cached.html"
        extra = tempdir/"extra.html"
        dummy = tempdir/"Slipbox.md"
        dummy.write_text(render_dummy(title), encoding="utf-8")
        script.write_text('\n'.join(generate_javascript(conn)),
                          encoding="utf-8")
        html.write_text('\n'.join(generate_active_htmls(conn)),
                        encoding="utf-8")
        with open(extra, "w", encoding="utf-8") as file:
            print(create_tag_pages(conn), file=file)
            print(create_tags(conn), file=file)
            print(create_reference_pages(conn), file=file)
            print(create_bibliography(conn), file=file)
        cmd = """{pandoc} Slipbox.md -Hscript.js --metadata title:{title} -Anav.html
                -Acached.html -Aextra.html -A{search} --section-divs {opts}
                -o {output} -c style.css
            """.format(
            pandoc=pandoc(),
            title=shlex.quote(title),
            opts=options,
            output=out/"index.html",
            search=data_shell_path("search.html"))
        subprocess.run(shlex.split(cmd), check=False, cwd=tempdir)
